import {
    createRequest,
    RequestError,
} from './xml-request';

import {
    buildBody,
    isAttrChanged,
    findDeviceById,
} from './utils';

import Log from './log';
import * as Parser from './xml-parser';

const TAG = 'api';

class CloudApi {

    _serverURL;
    _appId;
    _userName;
    _userId;
    _secToken;
    _devices = [];
    _apiListener;

    constructor(serverURL, appId) {
        this._serverURL = serverURL;
        this._appId = appId;
    }

    setApiListener(l) {
        this._apiListener = l;
    }

    setExtras(extras) {
        if (extras) {
            this._userName = extras.name;
            this._userId = extras.userId;
            this._secToken = extras.securityToken;
            this._devices = extras.devices;
        }
    }

}

class AsyncApi extends CloudApi {

    _retcast(response, callback, errorMsg, name, data) {
        let retCode = Parser.parserRetCode(response);
        if (retCode === '0') {
            if (name) {
                this._apiListener(name, data);
            }
            callback.onSuccess();
        } else {
            callback.onFail({code: retCode, msg: errorMsg});
        }
    }

    _loadDevAttrs(devices, callback, counter) {
        devices.forEach((device) => {
            let devCache = findDeviceById(this._devices, device.devId);
            if (devCache !== null && !isAttrChanged(device, devCache)) {
                this._loadDevAttrFromLocal(device, devCache, callback, counter);
            } else {
                this._loadDevAttrFromCloud(device, callback, counter);
            }
        });
    }

    _loadDevAttrFromLocal(device, devCache, callback, counter) {
        Log.info(TAG, 'loadDevAttrFromLocal() - ' + device.devId);
        device.typeName = devCache.typeName;
        device.attrs = devCache.attrs;
        counter.count++;
        let childrenLength = device.children.length;
        if (childrenLength !== 0) {
            counter.ok += childrenLength;
            this._loadDevAttrs(device.children, callback, counter);
        }
        if (counter.count === counter.ok) {
            callback();
        }
    }

    _loadDevAttrFromCloud(device, callback, counter) {
        Log.info(TAG, 'loadDevAttrFromCloud() - ' + device.devId);
        this.getDeviceAttributesWithValues(device.devId, device.typeId, {
            onSuccess: (result) => {
                Log.info(TAG, 'loadDevAttrFromCloud() - ' + device.devId + '/success');
                device.typeName = result.typeName;
                device.attrs = result.attrs;
                counter.count++;
                let childrenLength = device.children.length;
                if (childrenLength !== 0) {
                    counter.ok += childrenLength;
                    this._loadDevAttrs(device.children, callback, counter);
                }
                if (counter.count === counter.ok) callback();
            },
            onFail: (error) => {
                Log.error(TAG, 'loadDevAttrFromCloud() - ' + device.devId + '/failed');
                Log.error(TAG, error);
                counter.count++;
                let childrenLength = device.children.length;
                if (childrenLength !== 0) {
                    counter.ok += childrenLength;
                    this._loadDevAttrs(device.children, callback, counter);
                }
                if (counter.count === counter.ok) callback();
            }
        });
    }

    systemLogin(name, password, callback) {
        let request = createRequest(this._serverURL + 'systemLogin', 'POST', buildBody({
            appId: this._appId,
            name,
            password,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let result = Parser.parserSystemLogin(response);
                this._appId = result.appId;
                this._userId = result.userId;
                this._secToken = result.sysSecToken;
                this._userName = name;
                this._apiListener('systemLogin', {...result, name, password});
                callback.onSuccess(result);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    createDeviceType(typeName, displayName, callback) {
        let request = createRequest(this._serverURL + 'createDeviceType', 'POST', buildBody({
            sysSecToken: this._secToken,
            typeName,
            displayName,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'createDeviceType failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    createDeviceAttribute(type, name, device, ts, tsValueType, displayName, callback) {
        let request = createRequest(this._serverURL + 'createDeviceAttribute', 'POST', buildBody({
            sysSecToken: this._secToken,
            type,
            name,
            device,
            ts,
            tsValueType,
            displayName,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'createDeviceAttribute failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    registerDevice(pwd, deviceKey, name, callback) {
        let request = createRequest(this._serverURL + 'registerDevice', 'POST', buildBody({
            secToken: this._secToken,
            pwd,
            deviceKey,
            name,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'registerDevice failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    setJpushCredential(appKey, masterSec, callback) {
        let request = createRequest(this._serverURL + 'setJpushCredential', 'POST', buildBody({
            sysSecToken: this._secToken,
            appKey,
            masterSec,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'setJpushCredential failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    setAlismsCredential(accessKey, accessSecret, signName, registerationTemplate, recoverTemplate, callback) {
        let request = createRequest(this._serverURL + 'setAlismsCredential', 'POST', buildBody({
            sysSecToken: this._secToken,
            accessKey,
            accessSecret,
            signName,
            registerationTemplate,
            recoverTemplate,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'setAlismsCredential failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getSyslog(serviceId, category, start, end, skip, callback) {
        let request = createRequest(this._serverURL + 'getSyslog', 'POST', buildBody({
            sysSecToken: this._secToken,
            serviceId,
            category,
            start,
            end,
            skip,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let logs = Parser.parserSyslog(response);
                callback.onSuccess(logs);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getSyslogCount(serviceId, category, start, end, skip, callback) {
        let request = createRequest(this._serverURL + 'getSyslogCount', 'POST', buildBody({
            sysSecToken: this._secToken,
            serviceId,
            category,
            start,
            end,
            skip,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let count = Parser.parserSyslogCount(response);
                callback.onSuccess(count);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    userLogin(name, password, callback) {
        let request = createRequest(this._serverURL + 'userLogin', 'POST', buildBody({
            appId: this._appId,
            name,
            password,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let result = Parser.parserUserLogin(response);
                this._userId = result.userId;
                this._secToken = result.securityToken;
                this._apiListener('userLogin', {...result, name, password});
                callback.onSuccess();
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    updatePassword(password, newPassword, callback) {
        let request = createRequest(this._serverURL + 'updatePassword', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            password,
            newPassword,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'updatePassword failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    registerUser(code, phone, userName, password, callback) {
        let request = createRequest(this._serverURL + 'registerUser', 'POST', buildBody({
            appId: this._appId,
            code,
            phone,
            userName,
            password,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let userId = Parser.parserRegisterUser(response);
                callback.onSuccess(userId);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getVerificationCode(phone, type, callback) {
        let request = createRequest(this._serverURL + 'getVerificationCode', 'POST', buildBody({
            appId: this._appId,
            phone,
            type,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'getVerificationCode failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    recoverUserPassword(code, phone, password, callback) {
        let request = createRequest(this._serverURL + 'recoverUserPassword', 'POST', buildBody({
            appId: this._appId,
            code,
            phone,
            password,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'recoverUserPassword failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getUserValueList(callback) {
        let request = createRequest(this._serverURL + 'getUserValueList', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            userName: this._userName,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let values = Parser.parserUserValueList(response);
                this._apiListener('getUserValueList', values);
                callback.onSuccess(values);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    setUserAttribute(name, value, callback) {
        let request = createRequest(this._serverURL + 'setUserAttribute', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            name,
            value,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'setUserAttribute failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getDeviceTypeList(callback) {
        let request = createRequest(this._serverURL + 'getDeviceTypeList', 'POST', buildBody({
            secToken: this._secToken,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let types = Parser.parserDeviceTypeList(response);
                callback.onSuccess(types);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getDeviceList(callback) {
        let request = createRequest(this._serverURL + 'getDeviceList', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let devices = Parser.parserDeviceList(response);
                if (devices.length === 0) {
                    //数量为0的情况直接抛出了
                    this._apiListener('getDeviceList', devices);
                    callback.onSuccess(devices);
                } else {
                    //否则加载好所有属性
                    let counter = {ok: devices.length, count: 0};
                    this._loadDevAttrs(devices, () => {
                        this._apiListener('getDeviceList', devices);
                        callback.onSuccess(devices);
                    }, counter);
                }
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    createVirtualDevice(typeId, typeName, parentId, attributes, callback, flag = true) {
        let request = createRequest(this._serverURL + 'createVirtualDevice', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            typeId,
            typeName,
            parentId,
            attributes,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let deviceId = Parser.parserCreateVirtualDevice(response);
                if (flag) {
                    this._apiListener('createVirtualDevice', deviceId);
                }
                callback.onSuccess(deviceId);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getDeviceAttributesWithValues(devId, deviceTypeId, callback) {
        let request = createRequest(this._serverURL + 'getDeviceAttributesWithValues', 'POST', buildBody({
            secToken: this._secToken,
            devId,
            deviceTypeId,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let result = Parser.parserDeviceAttributesWithValues(response);
                callback.onSuccess(result);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getDeviceAttribute(devId, name, callback) {
        let request = createRequest(this._serverURL + 'getDeviceAttribute', 'POST', buildBody({
            secToken: this._secToken,
            devId,
            name,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let result = Parser.parserDeviceAttribute(response);
                callback.onSuccess(result);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    setDeviceAttribute(devId, typeId, name, value, callback) {
        let request = createRequest(this._serverURL + 'setDeviceAttribute', 'POST', buildBody({
            secToken: this._secToken,
            devId,
            typeId,
            name,
            value,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'setDeviceAttribute failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    setMultiDeviceAttributes(devId, typeId, attributes, callback) {
        let request = createRequest(this._serverURL + 'setMultiDeviceAttributes', 'POST', buildBody({
            secToken: this._secToken,
            devId,
            typeId,
            attributes,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'setMultiDeviceAttributes failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getDeviceTS2(devId, propName, start, end, callback) {
        let request = createRequest(this._serverURL + 'getDeviceTS2', 'POST', buildBody({
            secToken: this._secToken,
            devId,
            propName,
            start,
            end,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let result = Parser.parserDeviceTS2(response);
                callback.onSuccess(result);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getDeviceTS3(idArray, nameArray, start, end, skip, callback) {
        let request = createRequest(this._serverURL + 'getDeviceTS3', 'POST', buildBody({
            secToken: this._secToken,
            idArray,
            nameArray,
            start,
            end,
            skip,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let result = Parser.parserDeviceTS3(response);
                callback.onSuccess(result);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getDeviceTS3Count(idArray, nameArray, start, end, skip, callback) {
        let request = createRequest(this._serverURL + 'getDeviceTS3Count', 'POST', buildBody({
            secToken: this._secToken,
            idArray,
            nameArray,
            start,
            end,
            skip,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let count = Parser.parserDeviceTS3Count(response);
                callback.onSuccess(count);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    removeDeviceFromUser(deviceId, parentId, callback, flag = true) {
        let requestName = 'removeDeviceFromUser';
        let request = createRequest(this._serverURL + requestName, 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            deviceId,
            parentId,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let errorMsg = requestName + ' failed.';
                if (flag) {
                    this._retcast(response, callback, errorMsg, requestName, deviceId);
                } else {
                    this._retcast(response, callback, errorMsg);
                }
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    deviceAuth(name, password, callback) {
        let request = createRequest(this._serverURL + 'deviceAuth', 'POST', buildBody({
            secToken: this._secToken,
            name,
            password,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let devId = Parser.parserDeviceAuth(response);
                callback.onSuccess(devId);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    addDeviceToUser(deviceId, deviceName, typeName, parentId, callback, flag = true) {
        let requestName = 'addDeviceToUser';
        let request = createRequest(this._serverURL + requestName, 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            deviceId,
            deviceName,
            typeName,
            parentId,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let errorMsg = requestName + ' failed.';
                if (flag) {
                    this._retcast(response, callback, errorMsg, requestName, deviceId);
                } else {
                    this._retcast(response, callback, errorMsg);
                }
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    setHistoryAttached(id, attached, callback) {
        let request = createRequest(this._serverURL + 'setHistoryAttached', 'POST', buildBody({
            secToken: this._secToken,
            id,
            attached,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'setHistoryAttached failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    shareDevice(deviceId, targetUserId, ownership, callback) {
        let request = createRequest(this._serverURL + 'shareDevice', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            deviceId,
            targetUserId,
            ownership,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'shareDevice failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    cancelShareDevice(deviceId, targetUserId, callback) {
        let request = createRequest(this._serverURL + 'cancelShareDevice', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            deviceId,
            targetUserId,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'cancelShareDevice failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getDeviceUserList(devId, callback) {
        let request = createRequest(this._serverURL + 'getDeviceUserList', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            devId,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let users = Parser.parserDeviceUserList(response);
                callback.onSuccess(users);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    touchUser(targetUserId, msg, callback) {
        let request = createRequest(this._serverURL + 'touchUser', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            targetUserId,
            msg,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'touchUser failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getTouchList(callback) {
        let request = createRequest(this._serverURL + 'getTouchList', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let touchs = Parser.parserTouchList(response);
                callback.onSuccess(touchs);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    dealTouch(targetUserId, accept, callback) {
        let request = createRequest(this._serverURL + 'dealTouch', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            targetUserId,
            accept,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'dealTouch failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getUserInfoByName(userName, callback) {
        let request = createRequest(this._serverURL + 'getUserInfoByName', 'POST', buildBody({
            secToken: this._secToken,
            userName
        }));
        request.enqueue({
            onSuccess: (response) => {
                let user = Parser.parserUserInfoByName(response);
                callback.onSuccess(user);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    getContactsList(callback) {
        let request = createRequest(this._serverURL + 'getContactsList', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
        }));
        request.enqueue({
            onSuccess: (response) => {
                let contacts = Parser.parserContactsList(response);
                callback.onSuccess(contacts);
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    removeContact(targetUserId, callback) {
        let request = createRequest(this._serverURL + 'removeContact', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            targetUserId,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'removeContact failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

    checkVerificationCode(phone, code, type, callback) {
        let request = createRequest(this._serverURL + 'checkVerificationCode', 'POST', buildBody({
            appId: this._appId,
            phone,
            code,
            type,
        }));
        request.enqueue({
            onSuccess: (response) => {
                this._retcast(response, callback, 'checkVerificationCode failed.');
            },
            onFail: (error) => {
                callback.onFail(error);
            }
        });
    }

}

/********************************************************************************************************/

class SyncApi extends CloudApi {

    async _repexe(request, pf, name) {
        let response = await request.execute();
        let data = pf(response);
        if (name) {
            this._apiListener(name, data);
        }
        return data;
    }

    async _retexe(request, name, data) {
        let retCode = await this._repexe(request, Parser.parserRetCode);
        if (retCode === '0') {
            if (name) {
                this._apiListener(name, data);
            }
            return true;
        } else {
            new ApiError(retCode, request._url + 'failed.');
        }
    }

    async _loadDevAttrs(devices) {
        for (let i = 0; i < devices.length; i++) {
            let device = devices[i];
            let devCache = findDeviceById(this._devices, device.devId);
            if (devCache !== null && !isAttrChanged(device, devCache)) {
                Log.info(TAG, 'loadDevAttrs() - ' + device.devId + '/local');
                device.typeName = devCache.typeName;
                device.attrs = devCache.attrs;
            } else {
                try {
                    let attrInfo = await this.getDeviceAttributesWithValues(device.devId, device.typeId);
                    Log.info(TAG, 'loadDevAttrs(): ' + device.devId + '/cloud');
                    device.typeName = attrInfo.typeName;
                    device.attrs = attrInfo.attrs;
                } catch (error) {
                    Log.error(TAG, 'loadDevAttrs(): ' + device.devId + '/cloud');
                    Log.error(TAG, error);
                }
            }
            let childrenLength = device.children.length;
            if (childrenLength !== 0) {
                this._loadDevAttrs(device.children);
            }
        }
    }

    async systemLogin(name, password) {
        let request = createRequest(this._serverURL + 'systemLogin', 'POST', buildBody({
            appId: this._appId,
            name,
            password,
        }));
        let response = await request.execute();
        let result = Parser.parserSystemLogin(response);
        this._appId = result.appId;
        this._userId = result.userId;
        this._secToken = result.sysSecToken;
        this._userName = name;
        this._apiListener('systemLogin', {...result, name, password});
        return true;
    }

    async createDeviceType(typeName, displayName) {
        return await this._retexe(createRequest(this._serverURL + 'createDeviceType', 'POST', buildBody({
            sysSecToken: this._secToken,
            typeName,
            displayName,
        })));
    }

    async createDeviceAttribute(type, name, device, ts, tsValueType, displayName) {
        return await this._retexe(createRequest(this._serverURL + 'createDeviceAttribute', 'POST', buildBody({
            sysSecToken: this._secToken,
            type,
            name,
            device,
            ts,
            tsValueType,
            displayName,
        })));
    }

    async registerDevice(pwd, deviceKey, name) {
        return await this._retexe(createRequest(this._serverURL + 'registerDevice', 'POST', buildBody({
            secToken: this._secToken,
            pwd,
            deviceKey,
            name,
        })));
    }

    async setJpushCredential(appKey, masterSec) {
        return await this._retexe(createRequest(this._serverURL + 'setJpushCredential', 'POST', buildBody({
            sysSecToken: this._secToken,
            appKey,
            masterSec,
        })));
    }

    async setAlismsCredential(accessKey, accessSecret, signName, registerationTemplate, recoverTemplate) {
        return await this._retexe(createRequest(this._serverURL + 'setAlismsCredential', 'POST', buildBody({
            sysSecToken: this._secToken,
            accessKey,
            accessSecret,
            signName,
            registerationTemplate,
            recoverTemplate,
        })));
    }

    async getSyslog(serviceId, category, start, end, skip) {
        return await this._repexe(createRequest(this._serverURL + 'getSyslog', 'POST', buildBody({
            sysSecToken: this._secToken,
            serviceId,
            category,
            start,
            end,
            skip,
        })), Parser.parserSyslog);
    }

    async getSyslogCount(serviceId, category, start, end, skip) {
        return await this._repexe(createRequest(this._serverURL + 'getSyslogCount', 'POST', buildBody({
            sysSecToken: this._secToken,
            serviceId,
            category,
            start,
            end,
            skip,
        })), Parser.parserSyslogCount);
    }

    async userLogin(name, password) {
        let request = createRequest(this._serverURL + 'userLogin', 'POST', buildBody({
            appId: this._appId,
            name,
            password,
        }));
        let response = await request.execute();
        let result = Parser.parserUserLogin(response);
        this._userId = result.userId;
        this._secToken = result.securityToken;
        this._apiListener('userLogin', {...result, name, password});
        return true;
    }

    async updatePassword(password, newPassword) {
        return await this._retexe(createRequest(this._serverURL + 'updatePassword', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            password,
            newPassword,
        })));
    }

    async registerUser(code, phone, userName, password) {
        return await this._repexe(createRequest(this._serverURL + 'registerUser', 'POST', buildBody({
            appId: this._appId,
            code,
            phone,
            userName,
            password,
        })), Parser.parserRegisterUser);
    }

    async getVerificationCode(phone, type) {
        return await this._retexe(createRequest(this._serverURL + 'getVerificationCode', 'POST', buildBody({
            appId: this._appId,
            phone,
            type,
        })));
    }

    async recoverUserPassword(code, phone, password) {
        return await this._retexe(createRequest(this._serverURL + 'recoverUserPassword', 'POST', buildBody({
            appId: this._appId,
            code,
            phone,
            password,
        })));
    }

    async getUserValueList() {
        return await this._repexe(createRequest(this._serverURL + 'getUserValueList', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            userName: this._userName,
        })), Parser.parserUserValueList, 'getUserValueList');
    }

    async setUserAttribute(name, value) {
        return await this._retexe(createRequest(this._serverURL + 'setUserAttribute', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            name,
            value,
        })));
    }

    async getDeviceTypeList() {
        return await this._repexe(createRequest(this._serverURL + 'getDeviceTypeList', 'POST', buildBody({
            secToken: this._secToken,
        })), Parser.parserDeviceTypeList);
    }

    async getDeviceList() {
        let request = createRequest(this._serverURL + 'getDeviceList', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
        }));
        let response = await request.execute();
        let devices = Parser.parserDeviceList(response);
        if (devices.length === 0) {
            this._apiListener('getDeviceList', devices);
        } else {
            this._loadDevAttrs(devices);
        }
        return devices;
    }

    async createVirtualDevice(typeId, typeName, parentId, attributes) {
        return await this._repexe(createRequest(this._serverURL + 'createVirtualDevice', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            typeId,
            typeName,
            parentId,
            attributes,
        })), Parser.parserCreateVirtualDevice, 'createVirtualDevice');
    }

    async getDeviceAttributesWithValues(devId, deviceTypeId) {
        return await this._repexe(createRequest(this._serverURL + 'getDeviceAttributesWithValues', 'POST', buildBody({
            secToken: this._secToken,
            devId,
            deviceTypeId,
        })), Parser.parserDeviceAttributesWithValues);
    }

    async getDeviceAttribute(devId, name) {
        return await this._repexe(createRequest(this._serverURL + 'getDeviceAttribute', 'POST', buildBody({
            secToken: this._secToken,
            devId,
            name,
        })), Parser.parserDeviceAttribute);
    }

    async setDeviceAttribute(devId, typeId, name, value) {
        return await this._retexe(createRequest(this._serverURL + 'setDeviceAttribute', 'POST', buildBody({
            secToken: this._secToken,
            devId,
            typeId,
            name,
            value,
        })));
    }

    async setMultiDeviceAttributes(devId, typeId, attributes) {
        return await this._retexe(createRequest(this._serverURL + 'setMultiDeviceAttributes', 'POST', buildBody({
            secToken: this._secToken,
            devId,
            typeId,
            attributes,
        })));
    }

    async getDeviceTS2(devId, propName, start, end) {
        return await this._repexe(createRequest(this._serverURL + 'getDeviceTS2', 'POST', buildBody({
            secToken: this._secToken,
            devId,
            propName,
            start,
            end,
        })), Parser.parserDeviceTS2);
    }

    async getDeviceTS3(idArray, nameArray, start, end, skip) {
        return await this._repexe(createRequest(this._serverURL + 'getDeviceTS3', 'POST', buildBody({
            secToken: this._secToken,
            idArray,
            nameArray,
            start,
            end,
            skip,
        })), Parser.parserDeviceTS3);
    }

    async getDeviceTS3Count(idArray, nameArray, start, end, skip) {
        return await this._repexe(createRequest(this._serverURL + 'getDeviceTS3Count', 'POST', buildBody({
            secToken: this._secToken,
            idArray,
            nameArray,
            start,
            end,
            skip,
        })), Parser.parserDeviceTS3Count);
    }

    async removeDeviceFromUser(deviceId, parentId) {
        return await this._retexe(createRequest(this._serverURL + 'removeDeviceFromUser', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            deviceId,
            parentId,
        })), 'removeDeviceFromUser', deviceId);
    }

    async deviceAuth(name, password) {
        return await this._repexe(createRequest(this._serverURL + 'deviceAuth', 'POST', buildBody({
            secToken: this._secToken,
            name,
            password,
        })), Parser.parserDeviceAuth);
    }

    async addDeviceToUser(deviceId, deviceName, typeName, parentId) {
        return await this._retexe(createRequest(this._serverURL + 'addDeviceToUser', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            deviceId,
            deviceName,
            typeName,
            parentId,
        })), 'addDeviceToUser', deviceId);
    }

    async setHistoryAttached(id, attached) {
        return await this._retexe(createRequest(this._serverURL + 'setHistoryAttached', 'POST', buildBody({
            secToken: this._secToken,
            id,
            attached,
        })));
    }

    async shareDevice(deviceId, targetUserId, ownership) {
        return await this._retexe(createRequest(this._serverURL + 'shareDevice', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            deviceId,
            targetUserId,
            ownership,
        })));
    }

    async cancelShareDevice(deviceId, targetUserId) {
        return await this._retexe(createRequest(this._serverURL + 'cancelShareDevice', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            deviceId,
            targetUserId,
        })));
    }

    async getDeviceUserList(devId) {
        return await this._repexe(createRequest(this._serverURL + 'getDeviceUserList', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            devId,
        })), Parser.parserDeviceUserList, 'getDeviceUserList');
    }

    async touchUser(targetUserId, msg) {
        return await this._retexe(createRequest(this._serverURL + 'touchUser', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            targetUserId,
            msg,
        })));
    }

    async getTouchList() {
        return await this._repexe(createRequest(this._serverURL + 'getTouchList', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
        })), Parser.parserTouchList);
    }

    async dealTouch(targetUserId, accept) {
        return await this._retexe(createRequest(this._serverURL + 'dealTouch', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            targetUserId,
            accept,
        })));
    }

    async getUserInfoByName(userName) {
        return this._repexe(createRequest(this._serverURL + 'getUserInfoByName', 'POST', buildBody({
            secToken: this._secToken,
            userName
        })), Parser.parserUserInfoByName);
    }

    async getContactsList() {
        return this._repexe(createRequest(this._serverURL + 'getContactsList', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
        })), Parser.parserContactsList);
    }

    async removeContact(targetUserId) {
        return await this._retexe(createRequest(this._serverURL + 'removeContact', 'POST', buildBody({
            secToken: this._secToken,
            userId: this._userId,
            targetUserId,
        })));
    }

    async checkVerificationCode(phone, code, type) {
        return await this._retexe(createRequest(this._serverURL + 'checkVerificationCode', 'POST', buildBody({
            appId: this._appId,
            phone,
            code,
            type,
        })));
    }

}

export class ApiError extends RequestError {

    constructor(code, msg) {
        super({code, msg});
    }

}

export function createApi(serverURL, appId, sync = false) {
    if (sync) {
        return new SyncApi(serverURL, appId);
    } else {
        return new AsyncApi(serverURL, appId);
    }
}