import Log from './log';
import TokenTimer from './token-timer';
import { createApi } from './api';
import { createMqtt } from './mqtt';
import { createDatabase } from './database';
import { updateDevice, findChange } from './utils';

const TAG = 'store';

class Store {
    
    _platform = '';

    _deviceTreeListeners = [];
    _deviceListeners = [];
    _deviceAttributeListeners = [];
    _deleteDeleteListeners = [];
    _addDeviceListeners = [];

    _tokenTimer = {};
    _topics = [];
    _mqtt = {};

    database = {};
    api = {};

    constructor(platform, config) {
        this._platform = platform;
        //创建API
        this.api = createApi(config.serverURL, config.appId);
        this.api.setApiListener(this._apiListener);

        //创建MQTT
        this._mqtt = createMqtt(platform, config.socketIP, config.socketPort);
        this._mqtt.setMessageListener(this._messageListener);
        this._mqtt.setCloseListener(this._closeListener);

        //创建数据库
        this.database = createDatabase(platform);

        //初始化token定时器
        this._tokenTimer = new TokenTimer(
            this.database,
            this.api,
        );

        //web端按F5刷新后store将重新创建，为了使之前的数据不丢失，在这里重新setup
        let user = this.database.load('user');
        if (user) {
            Log.info(TAG, 'setExtras() - user:');
            Log.info(TAG, user);
            this.api.setExtras(user);
        }
    }

    _apiListener = (apiName, apiData) => {
        if (apiName === 'systemLogin') {
            //管理员登录
            this.database.save('user', apiData);
            if (this._tokenTimer.isRun()) {
                this._tokenTimer.stop();
            }
            this._tokenTimer.start();
        } else if (apiName === 'userLogin') {
            //普通用户登录
            this.database.save('user', apiData);
            if (this._tokenTimer.isRun()) {
                this._tokenTimer.stop();
            }
            this._tokenTimer.start();
        } else if (apiName === 'getDeviceList') {
            //获取设备列表
            let dOldUser = this.database.load('user');
            let dNewUser = { ...dOldUser, devices: apiData };
            this.database.save('user', dNewUser);
            //判断是否连接MQTT
            if (this._mqtt.isConnected()) {
                //取消订阅，在此重新订阅
                this._mqtt.unsubscribe(this._topics);
                this._clearTopics();
                this._generateTopics(apiData);
                this._mqtt.subscribe(this._topics);
            } else {
                this._generateTopics(apiData);
                this.connectCloud();
            }
        } else if (apiName === 'getUserValueList') {
            //获取用户属性列表
            let uOldUser = this.database.load('user');
            let uNewUser = { ...uOldUser, attributes: apiData };
            this.database.save('user', uNewUser);
        } else if (apiName === 'addDeviceToUser') {
            //添加设备
            this._sendAddDeviceMessage(apiName);
        } else if (apiName === 'createVirtualDevice') {
            //创建设备
            this._sendAddDeviceMessage(apiName);
        } else if (apiName === 'removeDeviceFromUser') {
            //删除设备
            this._sendDeleteDeviceMessage(apiName);
        }
    }

    _messageListener = (topic, payloadString) => {
        let array = topic.split('/');
        if (array.lenght < 4) {
            Log.error(TAG, 'MessageListener{} - topic error.');
            return;
        }

        let devId = array[2];
        let name = array[3];
        let value = payloadString;
        let updTime = new Date().getTime();

        Log.info(TAG, 'MessageListener{} - ' + topic + '/' + payloadString);

        if (name === 'addDeviceResult') {
            if (value === '0') {
                this._sendAddDeviceMessage('addByGateway');
            } else {
                this._addDeviceListeners.forEach((listener) => {
                    listener({ code: value, msg: 'add fail.' }, undefined);
                });
            }
        } else if (name === 'deleteDeviceResult') {
            if (value === '0') {
                this._sendDeleteDeviceMessage('deleteFromGateway');
            } else {
                this._deleteDeleteListeners.forEach((listener) => {
                    listener({ code: value, msg: 'delete fail.' }, undefined);
                });
            }
        } else {
            let user = this.database.load('user');
            let devices = user.devices;
            updateDevice(devices, devId, name, value, updTime);
            this.database.save('user', { ...user, devices });
            this._deviceTreeListeners.forEach((item) => {
                item(devices);
            });
            this._deviceListeners.forEach((item) => {
                if (item.devId === devId) {
                    item.listener({
                        name: name,
                        value: value,
                        updTime: updTime,
                    });
                }
            });
            this._deviceAttributeListeners.forEach((item) => {
                if (item.devId === devId && item.name === name) {
                    item.listener({
                        value: value,
                        updTime: updTime,
                    });
                }
            });
        }
    }

    _closeListener = (error) => {
        //mqtt被关闭了
        Log.error(TAG, 'CloseListener{} - mqtt closed.');
        Log.error(TAG, error);
        if (this._platform === 'mobile') {
            this.connectCloud();
        }
    }

    _sendAddDeviceMessage(apiName) {
        Log.info(TAG, 'sendAddDeviceMessage() - start.');
        this.api.getDeviceList({
            onSuccess: (devices) => {
                Log.info(TAG, 'sendAddDeviceMessage() - success.');
                let user = this.database.load('user');
                let uDevices = user.devices;
                findChange(uDevices, devices, (device) => {
                    this._addDeviceListeners.forEach((listener) => {
                        listener(undefined, device);
                    });
                });
            },
            onFail: (error) => {
                Log.error(TAG, 'sendAddDeviceMessage() - failed.');
                Log.error(TAG, error);
                this._addDeviceListeners.forEach((listener) => {
                    listener(error, undefined);
                });
            }
        });
    }

    _sendDeleteDeviceMessage = function (apiName) {
        Log.info(TAG, '_sendDeleteDeviceMessage() - start.');
        this.api.getDeviceList({
            onSuccess: (devices) => {
                Log.info(TAG, '_sendDeleteDeviceMessage() - success.');
                let user = this.database.load('user');
                let uDevices = user.devices;
                findChange(devices, uDevices, (device) => {
                    this._deleteDeleteListeners.forEach((listener) => {
                        listener(undefined, device);
                    });
                });
            },
            onFail: (error) => {
                Log.error(TAG, '_sendDeleteDeviceMessage() - failed.');
                Log.error(TAG, error);
                this._deleteDeleteListeners.forEach((listener) => {
                    listener(error, undefined);
                });
            }
        });
    }

    _clearTopics() {
        let lenght = this._topics.length;
        this._topics.splice(0, lenght);
    }

    _generateTopics(devices) {
        devices.forEach((device) => {
            let topic = 'property0/' + device.typeId + '/' + device.devId + '/+';
            this._topics.push(topic);
            if (device.children.length !== 0) {
                this._generateTopics(device.children);
            }
        });
    }

    connectCloud() {
        Log.info(TAG, 'connectCloud(): start.');
        this._mqtt.connect(() => {
            Log.info(TAG, 'connectCloud() - success, subscribe size:' + this._topics.length);
            this._mqtt.subscribe(this._topics);
        }, (error) => {
            Log.error(TAG, 'connectCloud() - failed.');
            Log.error(TAG, error);
            setTimeout(() => {
                this.connectCloud();
            }, 3000);
        });
    }

    disconnectCloud() {
        Log.info(TAG, 'disconnectCloud() - start.');
        this._mqtt.unsubscribe(this._topics);
        this._mqtt.disconnect();
        Log.info(TAG, 'disconnectCloud() - success.');
    }

    destroy() {
        this.disconnectCloud();
        Log.info(TAG, 'destroy() - success.');
    }

    setDeviceTreeListener(listener) {
        this._deviceTreeListeners.push(listener);
        return () => {
            let deleteIndex = this._deviceTreeListeners.findIndex((item) => listener === item);
            if (deleteIndex !== -1) this._deviceTreeListeners.splice(deleteIndex, 1);
        };
    }

    setDeviceListener(devId, listener) {
        this._deviceListeners.push({ devId, listener });
        return () => {
            let deleteIndex = this._deviceListeners.findIndex((item) => listener === item.listener);
            if (deleteIndex !== -1) this._deviceListeners.splice(deleteIndex, 1);
        };
    }

    setDeviceAttributeListener(devId, name, listener) {
        this._deviceAttributeListeners.push({ devId, name, listener });
        return () => {
            let deleteIndex = this._deviceAttributeListeners.findIndex((item) => listener === item.listener);
            if (deleteIndex !== -1) this._deviceAttributeListeners.splice(deleteIndex, 1);
        };
    }

    setDeleteDeviceListener(listener) {
        this._deleteDeleteListeners.push(listener);
        return () => {
            let deleteIndex = this._deleteDeleteListeners.findIndex((item) => listener === item);
            if (deleteIndex !== -1) this._deleteDeleteListeners.splice(deleteIndex, 1)
        };
    }

    setAddDeviceListener(listener) {
        this._addDeviceListeners.push(listener);
        return () => {
            let deleteIndex = this._addDeviceListeners.findIndex((item) => listener === item);
            if (deleteIndex !== -1) this._addDeviceListeners.splice(deleteIndex, 1)
        };
    }
}

export default Store;