/*
 * File Name                   : XML请求文件
 * Author                      : Jason
 * Version                     : 1.0.0
 * Date                        : 2017/4/15
 * Revision History            : 
 * Copyright (circle) 2015-16 Axalent Solutions (Shanghai) Co., Ltd.
 * All rights reserved.
 */

import { DOMParser } from 'xmldom';
import { parserRequestError } from './xml-parser';

const NETWORK_FAILED_WEB = 'TypeError: Network request failed';
const NETWORK_FAILED_MOBILE = 'TypeError: Failed to fetchat TypeError (native)';

export class RequestError extends Error {

    code = -1;

    constructor(e) {
        super(e.msg);
        this.code = e.code;
    }
}

class Request {

    _url = null;
    _method = null;
    _body = null;

    constructor(url, method, body) {
        this._url = url;
        this._method = method;
        this._body = body;
    }

    execute() {
    }

    enqueue(callback) {
    }

}

class FetchRequest extends Request {

    init = null;

    constructor(url, method, body) {
        super(url, method, body);
        this.init = {
            body: this._body,
            method: this._method,
            headers: { 'Content-type': 'application/x-www-form-urlencoded' },
        };
    }

    _toXML(text) {
        let parser = new DOMParser();
        let doc = parser.parseFromString(text, 'text/xml');
        return doc;
    }
    
    _parserFetchError(error) {
        let msg = error.toString();
        if (msg === NETWORK_FAILED_WEB || msg === NETWORK_FAILED_MOBILE) {
            return { code: -2, msg: 'network error.' };
        } else {
            return { code: -1, msg: 'unknown error.' };
        }
    }

    async execute() {
        try {
            let response = await fetch(this._url, this.init);
            let text = await response.text();
            let doc = this._toXML(text);
            if (response.ok) {
                return doc;
            } else {
                new RequestError(parserRequestError(doc));
            }
        } catch (error) {
            new RequestError(this._parserFetchError(error));
        }
    }

    enqueue(callback) {
        let ok = false;
        fetch(this._url, this.init)
        .then((response) => {
            ok = response.ok;
            return response.text();
        })
        .then((text) => {
            let doc = this._toXML(text);
            if (ok) {
                callback.onSuccess(doc);
            } else {
                callback.onFail(parserRequestError(doc));
            }
        })
        .catch((error) => {
            callback.onFail(this._parserFetchError(error));
        });
    }

}

class AjaxRequest extends Request{

    execute() {
        // no support
    }

    enqueue(callback) {
        let request = new XMLHttpRequest();
        request.timeout = 30000;
        request.ontimeout = () => {
            callback.onFail({ code: -2 , msg: 'request timeout.' });
        };
        request.onreadystatechange = (e) => {
            if (request.readyState !== 4) {
                return;
            }
            let doc = request.responseXML;
            if (request.status === 200) {
                callback.onSuccess(doc);
            } else {
                callback.onFail({ code: request.status, msg: 'request error.' });
            }
        };
        request.open(this._method, this._url);
        request.setRequestHeader('Content-type', 'application/x-www-form-urlencoded');
        request.send(this._body);
    }

}

export function createRequest(url, method, body, type = 'fetch') {
    if (type === 'fetch') {
        return new FetchRequest(url, method, body);
    } else {
        return new AjaxRequest(url, method, body);
    }
}